/*==============================================================================
 The Fraclet Programming Model - Fractal Component Model (fractal@objectweb.org)
 Copyright (C) 2002-2006 INRIA Futurs / LIFL
 Copyright (C) 2006-2007 Universitetet i Oslo

 This library is free software; you can redistribute it and/or modify it under 
 the terms of the GNU Lesser General Public License as published by the Free 
 Software Foundation; either version 2.1 of the License, or any later version.

 This library is distributed in the hope that it will be useful, but WITHOUT ANY 
 WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public License along 
 with this library; if not, write to the Free Software Foundation, Inc., 
 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA

 Initial developer(s): Romain Rouvoy
 Contributor(s): 
 =============================================================================*/
package org.objectweb.fractal.spoonlet.attribute;

import static java.util.logging.Level.FINER;
import static java.util.logging.Level.FINEST;
import static java.util.logging.Logger.getLogger;
import static org.objectweb.fractal.spoonlet.attribute.AttributeHelper.AC;
import static org.objectweb.fractal.spoonlet.attribute.AttributeHelper.attributeName;
import static org.objectweb.fractal.spoonlet.attribute.AttributeHelper.controllerFullname;
import static org.objectweb.fractal.spoonlet.component.ComponentHelper.hasFieldAnnotation;
import static org.objectweb.fractal.spoonlet.utils.Constants.LOG;
import static spoon.reflect.declaration.ModifierKind.PUBLIC;
import static spoon.template.Substitution.insertAll;

import java.util.HashSet;
import java.util.Set;
import java.util.logging.Logger;

import org.objectweb.fractal.fraclet.annotations.Attribute;

import spoon.processing.AbstractAnnotationProcessor;
import spoon.reflect.declaration.CtClass;
import spoon.reflect.declaration.CtField;
import spoon.reflect.declaration.CtInterface;
import spoon.reflect.declaration.ModifierKind;
import spoon.reflect.reference.CtTypeReference;
import spoon.template.Template;

/**
 * An annotation processor to manage the Fraclet annotation
 * <code>Attribute</code>.
 * 
 * @author Romain Rouvoy (Contributor)
 * @author Nicolas Pessemier (Initial Developer)
 * @version $Revision$
 */
public class AttributeProcessor extends
        AbstractAnnotationProcessor<Attribute, CtField<?>> {
    private final Logger log = getLogger("spoonlet.fractal.attribute");

    /* (non-Javadoc)
     * @see spoon.processing.AbstractProcessor#init()
     */
    @Override
    public void init() {
        addProcessedAnnotationType(Attribute.class);
        super.init();
    }

    /* (non-Javadoc)
     * @see spoon.processing.AbstractAnnotationProcessor#inferConsumedAnnotationType()
     */
    @Override
    public boolean inferConsumedAnnotationType() {
        return false;
    }

    /* (non-Javadoc)
     * @see spoon.processing.AnnotationProcessor#process(java.lang.annotation.Annotation,spoon.reflect.declaration.CtElement)
     */
    public void process(Attribute ann, CtField<?> fld) {
        CtClass<?> comp = fld.getParent(CtClass.class);
        if (LOG && log.isLoggable(FINER))
            log.finer("Fractal - Processing field " + comp.getSimpleName()
                    + "." + fld.getSimpleName() + "...");
        CtInterface<?> ac = getFactory().Interface().get(
                controllerFullname(comp));
        if (ac == null) // creation of the attribute controller interface
            ac = createAttributeController(comp);

        String attr = attributeName(fld, ann);
        if (LOG && log.isLoggable(FINEST))
            log.finest("Inserting getter/setters for attribute " + attr
                    + "...");
        Template tmpl = new GetterSetterTemplate(fld, attr);
        insertAll(ac, tmpl); // insert code into the controller
        insertAll(comp, tmpl); // insert code into the class
    }

    /**
     * Creates an attribute controller definition from a component definition.
     * 
     * @param comp the component describing the attributes.
     * @return the associated attribute controller.
     */
    private final CtInterface<?> createAttributeController(CtClass<?> comp) {
        String name = controllerFullname(comp);
        if (LOG && log.isLoggable(FINEST))
            log.finest("Creating attribute controller " + name + "...");
        CtInterface<?> ac = getFactory().Interface().create(name);
        Set<ModifierKind> modifiers = new HashSet<ModifierKind>();
        modifiers.add(PUBLIC);
        ac.setModifiers(modifiers);
        ac.getSuperInterfaces().add(superAttributeController(comp));
        comp.getSuperInterfaces().add(ac.getReference());
        return ac;
    }

    /**
     * Provides the reference of the super attribute controller class.
     * 
     * @param comp class containing the attributes.
     * @return the reference of the super attribute controller class.
     */
    private final CtTypeReference<?> superAttributeController(CtClass<?> comp) {
        CtTypeReference<?> parent = comp.getSuperclass();
        if (parent == null)
            return getFactory().Type().createReference(AC);
        CtClass<?> p = (CtClass<?>) parent.getDeclaration();
        if (hasFieldAnnotation(p, Attribute.class, false))
            return getFactory().Type().createReference(controllerFullname(p));
        return superAttributeController(p);
    }
}