package org.objectweb.fractal.spoonlet.binding;

import static java.util.logging.Level.FINER;
import static java.util.logging.Logger.getLogger;
import static org.objectweb.fractal.spoonlet.binding.BindingHelper.BC;
import static org.objectweb.fractal.spoonlet.component.ComponentHelper.hasController;
import static org.objectweb.fractal.spoonlet.component.ComponentHelper.hasFieldAnnotation;
import static org.objectweb.fractal.spoonlet.utils.Constants.LOG;
import static spoon.template.Substitution.insertAll;
import static spoon.template.Substitution.substituteMethodBody;

import java.lang.annotation.Annotation;
import java.util.ArrayList;
import java.util.List;
import java.util.logging.Logger;

import org.objectweb.fractal.fraclet.annotations.Requires;

import spoon.processing.AbstractAnnotationProcessor;
import spoon.reflect.code.CtStatement;
import spoon.reflect.declaration.CtClass;
import spoon.reflect.declaration.CtField;
import spoon.reflect.reference.CtTypeReference;
import spoon.template.Template;

/**
 * Abstract binding processor.
 * 
 * @author Romain Rouvoy (Initial Developer)
 * @version $Revision$
 * @param <A> Type of annotation processed by the binding processor.
 */
public abstract class AbstractBindingProcessor<A extends Annotation> extends
        AbstractAnnotationProcessor<A, CtField<?>> {
    private final Logger log = getLogger("spoonlet.fractal.binding");

    /* (non-Javadoc)
     * @see spoon.processing.AbstractAnnotationProcessor#inferConsumedAnnotationType()
     */
    @Override
    public boolean inferConsumedAnnotationType() {
        return false;
    }

    /* (non-Javadoc)
     * @see spoon.processing.AnnotationProcessor#process(java.lang.annotation.Annotation, spoon.reflect.declaration.CtElement)
     */
    public void process(A a, CtField<?> f) {
        CtClass<?> comp = f.getParent(CtClass.class);
        if (LOG && log.isLoggable(FINER))
            log.finer("Fractal - Processing field " + comp.getSimpleName()
                    + "." + f.getSimpleName() + "...");
        insertMethodSignatures(comp, f, a);
        insertMethodBodies(comp, f, a);
    }

    /**
     * Inserts the method signatures associated to the processed field.
     * 
     * @param c the component processed.
     * @param f the field processed.
     * @param a the annotation processed.
     */
    protected void insertMethodSignatures(CtClass<?> c, CtField<?> f, A a) {
        if (hasController(c, getFactory().Interface().createReference(BC),
                false))
            return; // Class already implements BindingController interface
        if (c.getSuperclass() == null) { // No parent class
            insertAll(c, new BindingControllerTemplate());
            return;
        }
        CtClass<?> parent = (CtClass<?>) c.getSuperclass().getDeclaration();
        if (!hasFieldAnnotation(parent, Requires.class, true))
            insertAll(c, new BindingControllerTemplate());
        else
            insertAll(c, new BindingExtensionTemplate());
    }

    /**
     * Insert a new body in the methods.
     * 
     * @param c reference to the class
     * @param f reference to the field
     * @param a reference to the annotation
     */
    protected abstract void insertMethodBodies(CtClass<?> c, CtField<?> f, A a);

    /**
     * Insert a piece of code in at the beginning of a method.
     * 
     * @param t template used for generating the piece of code to insert.
     * @param c target class for the insertion
     * @param m target method for the insertion
     * @param a method parameters
     * @return
     */
    protected void insertBegin(Template t, CtClass<?> c, String m,
            Class<?>... a) {
        List<CtTypeReference<?>> types = new ArrayList<CtTypeReference<?>>();
        for (Class<?> cls : a)
            types.add(getFactory().Type().createReference(cls));
        CtTypeReference<?>[] args = types.toArray(new CtTypeReference<?>[0]);
        CtStatement code = substituteMethodBody(c, t, m, args).getStatements()
                .get(0);
        c.getMethod(m, args).getBody().insertBegin(code);
    }
}