/*==============================================================================
 The Fraclet Programming Model - Fractal Component Model (fractal@objectweb.org)
 Copyright (C) 2002-2006 INRIA Futurs / LIFL
 Copyright (C) 2006-2007 Universitetet i Oslo

 This library is free software; you can redistribute it and/or modify it under 
 the terms of the GNU Lesser General Public License as published by the Free 
 Software Foundation; either version 2.1 of the License, or any later version.

 This library is distributed in the hope that it will be useful, but WITHOUT ANY 
 WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public License along 
 with this library; if not, write to the Free Software Foundation, Inc., 
 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA

 Initial developer(s): Nicolas Pessemier
 Contributor(s): Romain Rouvoy
 =============================================================================*/
package org.objectweb.fractal.spoonlet.lifecycle;

import static java.util.logging.Level.FINER;
import static java.util.logging.Logger.getLogger;
import static org.objectweb.fractal.spoonlet.component.ComponentHelper.hasController;
import static org.objectweb.fractal.spoonlet.component.ComponentHelper.hasFieldAnnotation;
import static org.objectweb.fractal.spoonlet.utils.Constants.LOG;
import static spoon.template.Substitution.insertAll;
import static spoon.template.Substitution.substituteMethodBody;

import java.util.ArrayList;
import java.util.List;
import java.util.logging.Logger;

import org.objectweb.fractal.api.control.LifeCycleController;
import org.objectweb.fractal.fraclet.annotations.Lifecycle;

import spoon.processing.AbstractAnnotationProcessor;
import spoon.reflect.declaration.CtClass;
import spoon.reflect.declaration.CtConstructor;
import spoon.reflect.declaration.CtMethod;
import spoon.reflect.reference.CtTypeReference;
import spoon.template.Template;

/**
 * An annotation processor to manage the Fraclet annotation
 * <code>Requires</code>.
 * 
 * @author Romain Rouvoy (Initial Developer)
 * @version $Revision$
 */
public class LifecycleProcessor extends
        AbstractAnnotationProcessor<Lifecycle, CtMethod<?>> {
    private final Logger log = getLogger("spoonlet.fractal.lifecycle");

    /** Constant referring to the lifecycle controller. */
    public static final Class<LifeCycleController> LCC = LifeCycleController.class;

    /* (non-Javadoc)
     * @see spoon.processing.AbstractProcessor#init()
     */
    @Override
    public void init() {
        addProcessedAnnotationType(Lifecycle.class);
        super.init();
    }

    /* (non-Javadoc)
     * @see spoon.processing.AnnotationProcessor#process(java.lang.annotation.Annotation, spoon.reflect.declaration.CtElement)
     */
    public void process(Lifecycle a, CtMethod<?> m) {
        CtClass<?> comp = m.getParent(CtClass.class);
        if (LOG && log.isLoggable(FINER))
            log.finer("Fractal - Processing method " + comp.getSimpleName()
                    + "." + m.getSimpleName() + "...");
        insertMethodSignatures(comp, m, a);
        insertMethodBodies(comp, m, a);
    }

    /**
     * @param c
     * @param f
     * @param a
     */
    protected void insertMethodSignatures(CtClass<?> c, CtMethod<?> f,
            Lifecycle a) {
        if (hasController(c, getFactory().Interface().createReference(
                LifeCycleController.class), false))
            return; // Class already implements BindingController interface
        if (c.getSuperclass() == null) { // No parent class
            CtConstructor<?> co = c.getConstructor();
            if (co.isImplicit())
                c.getConstructors().remove(co);
            insertAll(c, new LifecycleControllerTemplate(f));
            return;
        }
        CtClass<?> parent = (CtClass<?>) c.getSuperclass().getDeclaration();
        if (!hasFieldAnnotation(parent, Lifecycle.class, true)) {
            CtConstructor<?> co = c.getConstructor();
            if (co.isImplicit())
                c.getConstructors().remove(co);
            insertAll(c, new LifecycleControllerTemplate(f));
        } else
            insertAll(c, new LifecycleExtensionTemplate());
    }

    /**
     * Insert a new body in the methods.
     * 
     * @param c reference to the class
     * @param m reference to the field
     * @param a reference to the annotation
     */
    protected void insertMethodBodies(CtClass<?> c, CtMethod<?> m, Lifecycle a) {
        Template tmpl = new LifecycleTemplate(m);
        switch (a.value().value()) {
        case 0:
            insertEnd(tmpl, c, "createFc");
            break;
        case 1:
            insertEnd(tmpl, c, "startFc");
            break;
        case 2:
            insertEnd(tmpl, c, "stopFc");
            break;
        case 3:
            insertEnd(tmpl, c, "destroyFc");
            break;
        }
    }

    /**
     * Insert a piece of code in at the beginning of a method.
     * 
     * @param tmpl template used for generating the piece of code to insert.
     * @param comp target class for the insertion
     * @param method target method for the insertion
     * @param params method parameters
     */
    protected void insertEnd(Template tmpl, CtClass<?> comp, String method,
            Class<?>... params) {
        List<CtTypeReference<?>> types = new ArrayList<CtTypeReference<?>>();
        for (Class<?> cls : params)
            types.add(getFactory().Type().createReference(cls));
        CtTypeReference<?>[] args = types.toArray(new CtTypeReference<?>[0]);
        comp.getMethod(method, args).getBody().insertEnd(
                getFactory().Code().createStatementList(
                        substituteMethodBody(comp, tmpl, method, args)));
    }
}